#include "buffer.h"
#include <string.h>
#include <new.h>
#include <assert.h>

//--- Block -----------------------------------------------------------------
//The block contains lines separated by '\0'.
Block::Block()
  : bytesUsed(1),
    lines(1),
    c(new char[BLOCKSIZE])
{
        c[0]='\0';
}

Block::~Block() {
        delete[] c;
}

inline sbytenum_t Block::findLineStart(slinenum_t line) {
        return (sbytenum_t)(queryLinePtr(line)-c);
}

void Block::splitLine(slinenum_t line, colnum_t col) {
        sbytenum_t splitpos=findLineStart(line)+col;
        memmove(c+splitpos,c+splitpos+1,bytesUsed-splitpos);
        c[splitpos]='\0';
        bytesUsed++;
        lines++;
}

void Block::joinLine(slinenum_t line) {
        sbytenum_t joinpos=findLineStart(line+1);
        memmove(c+joinpos-1,c+joinpos,bytesUsed-joinpos);
        bytesUsed--;
        lines--;
}

void Block::removeLine(slinenum_t line) {
        sbytenum_t start=findLineStart(line);
        if(line<lines-1) {
                sbytenum_t nextstart=start;
                while(c[nextstart]!='\0') nextstart++;
                nextstart++;
                sbytenum_t ll = nextstart-start;
                memmove(c+start, c+nextstart, bytesUsed-nextstart);
                bytesUsed -= ll;
                lines--;
        } else {
                bytesUsed = start;
                lines--;
        }
}


char *Block::queryLinePtr(slinenum_t line) {
        char *p=c;
        while(line) {
                p = ((char*)memchr(p,'\0',BLOCKSIZE)) + 1;
                line--;
        }
        return p;
}

const char *Block::queryLinePtr(slinenum_t line) const {
        const char *p=c;
        while(line) {
                p = ((const char*)memchr(p,'\0',BLOCKSIZE)) + 1;
                line--;
        }
        return p;
}

inline sbytenum_t Block::queryLineLength(slinenum_t line) const {
        return (sbytenum_t)strlen(queryLinePtr(line));
}


void Block::splitTo(Block *bp) {
        sbytenum_t l,r;
        l=r=bytesUsed/2;
        sbytenum_t splitpos;
        for(;;) {
                if(c[l]=='\0') {
                        splitpos=l+1;
                        break;
                }
                if(c[r]=='\0') {
                        splitpos=r+1;
                        break;
                }
                l--;
                r++;
        }
        sbytenum_t splitbytes=bytesUsed-splitpos;
        memcpy(bp->c, c+splitpos, splitbytes);
        bp->bytesUsed=splitbytes;
        bp->recalcLines();

        bytesUsed=splitpos;
        recalcLines();
}

void Block::joinWith(const Block &bp) {
        memcpy(c+bytesUsed, bp.c, bp.bytesUsed);
        bytesUsed += bp.bytesUsed;
        lines += bp.lines;
}

void Block::recalcLines() {
        slinenum_t l=0;
	char *p=c;
        char *end=c+bytesUsed;
	while(p!=end) {
                p = ((char*)memchr(p,'\0',BLOCKSIZE)) + 1;
		l++;
	}
	
	lines=l;
}

void Block::insert(slinenum_t line, colnum_t col, char character) {
        sbytenum_t inspos=findLineStart(line)+col;
        memmove(c+inspos+1, c+inspos, bytesUsed-inspos);
        c[inspos]=character;
        bytesUsed++;
}

void Block::insert(slinenum_t line, colnum_t col, const char *s, sbytenum_t stringlength) {
        sbytenum_t inspos=findLineStart(line)+col;
        memmove(c+inspos+stringlength, c+inspos, bytesUsed-inspos);
        bytesUsed += stringlength;
        memcpy(c+inspos, s, stringlength);
}


void Block::remove(slinenum_t line, colnum_t col) {
        sbytenum_t delpos=findLineStart(line)+col;
        memmove(c+delpos, c+delpos+1, bytesUsed-delpos-1);
        bytesUsed--;
}

void Block::remove(slinenum_t line, colnum_t col, sbytenum_t chars) {
        sbytenum_t delpos=findLineStart(line)+col;
        memmove(c+delpos, c+delpos+chars, bytesUsed-delpos-chars);
        bytesUsed -= chars;
}

void Block::appendLine() {
        c[bytesUsed]='\0';
        lines++;
        bytesUsed++;
}


//--- Buffer ----------------------------------------------------------------
#define NOBLOCK ((unsigned)-1)

Buffer::Buffer() {
        blocksAlloc=16;
        blocks=0;
        block = (Block*)(new char[sizeof(Block)*blocksAlloc]);
        firstMark=0;
        
        changed=false;
        
        invalidateCachedFindLine();
}

Buffer::~Buffer() {
        for(unsigned b=0; b<blocks; b++) {
                block[b].~Block();
        }
        delete[] (char*)block;
}

inline void Buffer::fixcolpo(const Block *bp, slinenum_t line, colnum_t &col) {
        //limit col to the length of the line plus one
        if(col>0) {
                sbytenum_t ll=bp->queryLineLength(line);
                if(col>ll) col=ll;
        }
}

void Buffer::splitLine(llinenum_t line, colnum_t col) {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return;
        if(block[b].available()==0) {
                //split block
                splitBlock(b,line,rline);
        }
        //insert the character into the buffer
        fixcolpo(block+b,rline,col);
        block[b].splitLine(rline,col);

        changed=true;

        //update marks
        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                if(bm->queryLine()==line) {
                        if((bm->queryCol()==col && bm->stick==bm->right) ||
                           bm->queryCol()>col) 
                        {
                                bm->setLine(bm->queryLine()+1);
                                bm->setCol(bm->queryCol()-col);
                        }
                } else if(bm->queryLine()>line) {
                        bm->setLine(bm->queryLine()+1);
                }
        }
}

void Buffer::joinLine(llinenum_t line) {
        if(line==last_line) return;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return;
        sbytenum_t ll=block[b].queryLineLength(rline);
        if(rline==block[b].lines-1) {
		if(b==blocks-1) return; //cannot join last line in last block
                //joining across blocks
                //join with first line in next block
                if(block[b+1].queryLineLength(0)>block[b].available()) {
                        //block does not have enough room for first line in
			//next block - a split is necessary
                        splitBlock(b,line,rline);
                }
                block[b].insert(rline,ll,
                                block[b+1].queryLinePtr(0),block[b+1].queryLineLength(0));
                //now remove first line of the next block
                if(block[b+1].lines==1) {
                        //next block has only that line
                        removeBlock(b+1);
                } else
                        block[b+1].removeLine(0);
        } else {
                block[b].joinLine(rline);
        }
	
        changed=true;

        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                if(bm->queryLine()==line+1) {
                        bm->setLine(line);
                        bm->setCol(bm->queryCol()+ll);
                } else if(bm->queryLine()>line) {
                        bm->setLine(bm->queryLine()-1);
                }
        }
}

void Buffer::removeLine(llinenum_t line) {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return;
        if(block[b].lines==1) {
                //removing the last line in block removes the block too
                removeBlock(b);
        } else {
                block[b].removeLine(rline);
		if(block[b].lines==rline) invalidateCachedFindLine();
        }

        changed=true;

        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                if(bm->queryLine()==line) {
                        if(bm->stick==bm->left) {
                                bm->setLine(bm->queryLine()-1);
                                bm->setCol(queryLineLength(line));
                        } else {
                                bm->setCol(0);
                        }
                } else if(bm->queryLine()>line) {
                        bm->setLine(bm->queryLine()-1);
                }
        }
}

void Buffer::appendLine() {
        //if(blocks==0 || block[blocks-1]->available()==0) {
        if(blocks==0 || block[blocks-1].available()<block[blocks-1].bytesUsed/block[blocks-1].lines) {
                insertBlock(blocks);
        } else {
                block[blocks-1].appendLine();
        }
        
        changed=true;
}

void Buffer::insert(llinenum_t line, colnum_t col, char c) {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return;
        if(block[b].available()==0) {
                //split block
                splitBlock(b,line,rline);
        }
        //insert the character into the buffer
        fixcolpo(block+b,rline,col);
        block[b].insert(rline,col,c);

        changed=true;

        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                if(bm->queryLine()==line) {
                        if((bm->queryCol()==col && bm->stick==bm->right) ||
                           bm->queryCol()>col)
                                bm->setCol(bm->queryCol()+1);
                }
        }
}

void Buffer::insert(llinenum_t line, colnum_t col, const char *s, int sl) {
        if(line==last_line) line=queryLineCount()-1;
        sbytenum_t stringlength;
        if(sl==-1)
                stringlength=(sbytenum_t)strlen(s);
        else
                stringlength=sl;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return;
        if(block[b].available()<stringlength) {
                //split block
                splitBlock(b,line,rline);
        }
        fixcolpo(block+b,rline,col);
        block[b].insert(rline,col,s,stringlength);

        changed=true;

        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                if(bm->queryLine()==line) {
                        if((bm->queryCol()==col && bm->stick==bm->right) ||
                           bm->queryCol()>col)
                                bm->setCol(bm->queryCol()+stringlength);
                }
        }
}

void Buffer::remove(llinenum_t line, colnum_t col, unsigned chars) {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return;
        sbytenum_t ll=block[b].queryLineLength(rline);
        if(col<ll) {
                if(col+chars>ll)
                        block[b].remove(rline,col,ll-col);
                else
                        block[b].remove(rline,col,chars);
        }

        changed=true;

        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                if(bm->queryLine()==line) {
                        if(bm->queryCol()>=col) {
                                if(bm->queryCol()>=col+chars)
                                        bm->setCol(bm->queryCol()-chars);
                                else
                                        bm->setCol(col);
                        }
                }
        }
}

void Buffer::clear() {
        for(unsigned b=0; b<blocks; b++) {
                block[b].~Block();
        }
        delete[] (char*)block;

        blocksAlloc=16;
        blocks=0;
        block = (Block*)(new char[sizeof(Block)*blocksAlloc]);

        changed=true;
        invalidateCachedFindLine();

        for(BufferMark *bm=firstMark; bm; bm=bm->next) {
                bm->set(0,0);
        }
}


llinenum_t Buffer::queryLineCount() const {
        llinenum_t c=0;
        for(unsigned b=0; b<blocks; b++)
                c += block[b].lines;
        return c;
}

char *Buffer::queryLinePtr(llinenum_t line) {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return 0;
        return block[b].queryLinePtr(rline);
}

const char *Buffer::queryLinePtr(llinenum_t line) const {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return 0;
        return block[b].queryLinePtr(rline);
}

sbytenum_t Buffer::queryLineLength(llinenum_t line) const {
        if(line==last_line) line=queryLineCount()-1;
        slinenum_t rline;
        unsigned b=findBlockWithLine(line,&rline);
        if(b==NOBLOCK) return 0;
        return block[b].queryLineLength(rline);
}


unsigned Buffer::findBlockWithLine(llinenum_t line, slinenum_t *rline) const {
        if(cached_find_line==line) {
                *rline = (slinenum_t)(line - cached_find_line_line);
                return cached_find_line_block;
        }
        unsigned b;
        if(cached_find_line<=line) {
                llinenum_t tmp=line;
                line -= cached_find_line_line;
                cached_find_line = tmp;
                b = cached_find_line_block;
        } else {
                cached_find_line = line;
                cached_find_line_line = 0;
                cached_find_line_block = 0;
                b = 0;
        }
        
        llinenum_t cfll = cached_find_line_line;
        for(; b<blocks; b++) {
                slinenum_t block_lines=block[b].lines;
                if(block_lines<=line) {
                        line -= block_lines;
                        cfll += block_lines;
                } else {
                        *rline = (slinenum_t)line;
                        cached_find_line_block = b;
                        cached_find_line_line = cfll;
                        return b;
                }
        }

        invalidateCachedFindLine();
        return NOBLOCK;
}

void Buffer::splitBlock(unsigned &b, llinenum_t absoluteLine, slinenum_t &relativeLine) {
        //split block
        insertBlock(b+1);
        block[b].splitTo(block+b+1);
        if(cached_find_line_block>=b) invalidateCachedFindLine();
	if(relativeLine>=block[b].lines) {
		b++;
		relativeLine -= block[b].lines;
	}
}

void Buffer::insertBlock(unsigned before) {
        if(blocks==blocksAlloc) {
                //expand table
                unsigned nalloc=blocksAlloc+10;
                Block *newtbl=(Block*)(new char[sizeof(Block)*nalloc]);
                unsigned b;
                for(b=0; b<before; b++)
                        newtbl[b]=block[b];
                new(newtbl+before) Block;
                for(; b<blocks; b++)
                        newtbl[b+1]=block[b];
                delete[] (char*)block;
                block=newtbl;
                blocks++;
                blocksAlloc=nalloc;
        } else {
                for(unsigned b=blocks; b>before; b--)
                        block[b]=block[b-1];
                new(block+before) Block;
                blocks++;
        }
        if(cached_find_line_block>=before) invalidateCachedFindLine();
}

void Buffer::removeBlock(unsigned at) {
        block[at].~Block();
        for(unsigned b=at; b<blocks-1; b++)
                block[b]=block[b+1];
        blocks--;
        if(cached_find_line_block>=at) invalidateCachedFindLine();
}


void Buffer::addMark(BufferMark *bm) {
        bm->next=firstMark;
        firstMark=bm;
}

void Buffer::removeMark(BufferMark *bm) {
        if(bm==firstMark) {
                firstMark=bm->next;
                bm->next=0;
        } else {
                BufferMark *prev=firstMark;
                for(BufferMark *p=firstMark->next; p; prev=p, p=p->next) {
                        if(p==bm) {
                                prev->next=p->next;
                                p->next=0;
                                return;
                        }
                }
        }
}


void Buffer::compact() {
        for(unsigned b=0; b+1<blocks;) {
                if(block[b].available()>=block[b+1].bytesUsed) {
                        //join two blocks
                        block[b].joinWith(block[b+1]);
                        removeBlock(b+1);
                } else {
                        //move as much as possible from the next block to the current block
                        Block *b1=block+b,
                              *b2=block+b+1;
                        while(b1->available()>b2->queryLineLength(0)) {
                                sbytenum_t ll=b2->queryLineLength(0);
                                b1->appendLine();
                                b1->insert(b1->lines-1,0,
                                           b2->queryLinePtr(0),ll
                                          );
                                b2->removeLine(0);
                        }
                        b++;
			invalidateCachedFindLine();
                }
        }
}

void Buffer::fastCompact() {
        for(unsigned b=0; b+1<blocks; b++) {
                if(block[b].available()>=block[b+1].bytesUsed) {
                        //join two blocks
                        block[b].joinWith(block[b+1]);
                        removeBlock(b+1);
                } else {
			if(b>=blocks-3 ||
			   block[b+0].bytesUsed+block[b+1].bytesUsed+block[b+2].bytesUsed+block[b+3].bytesUsed > BLOCKSIZE*3)
				continue;
                        //move a few lines from the next block to the current block
                        Block *b1=block+b,
                              *b2=block+b+1;
                        sbytenum_t bytesMoved=0;
                        while(bytesMoved<BLOCKSIZE/64 && b1->available()>b2->queryLineLength(0)) {
                                sbytenum_t ll=b2->queryLineLength(0);
                                b1->appendLine();
                                b1->insert(b1->lines-1,0,
                                           b2->queryLinePtr(0),ll
                                          );
                                b2->removeLine(0);
                                bytesMoved += ll+1;
                        }
			invalidateCachedFindLine();
                }
        }
}


void Buffer::queryStatistics(lbytenum_t *used_bytes, lbytenum_t *allocated_bytes, llinenum_t *lines) {
        lbytenum_t ub=0;
        lbytenum_t ab=0;
        llinenum_t l=0;
        for(unsigned b=0; b<blocks; b++) {
                ub += block[b].bytesUsed;
                ab += BLOCKSIZE;
                l += block[b].lines;
        }
        *used_bytes = ub;
        *allocated_bytes = ab;
        *lines = l;
}
