#ifndef _BUFFER_H_
#define _BUFFER_H_

#include <limits.h>
#include <stddef.h>

//typedef unsigned short slinenum_t;
typedef size_t slinenum_t;

//typedef unsigned llinenum_t;
typedef size_t llinenum_t;

//typedef unsigned colnum_t;
typedef size_t colnum_t;

//typedef unsigned short sbytenum_t;
typedef size_t sbytenum_t;

//typedef unsigned lbytenum_t;
typedef size_t lbytenum_t;

#ifndef BLOCKSIZE
#define BLOCKSIZE 4096
#endif

struct Block {
        sbytenum_t bytesUsed;
        slinenum_t lines;
        char *c;
        
        Block();
        ~Block();
        
        void splitLine(slinenum_t line, colnum_t col);
        void joinLine(slinenum_t line);
        void removeLine(slinenum_t line);
        
        char *queryLinePtr(slinenum_t line);
        const char *queryLinePtr(slinenum_t line) const;
        sbytenum_t queryLineLength(slinenum_t line) const;
        sbytenum_t available() const { return (sbytenum_t)(BLOCKSIZE-bytesUsed); }
        void splitTo(Block *bp);
        void joinWith(const Block &bp);
        void insert(slinenum_t line, colnum_t col, char c);
        void insert(slinenum_t line, colnum_t col, const char *s, sbytenum_t stringlength);
        void remove(slinenum_t line, colnum_t col);
        void remove(slinenum_t line, colnum_t col, sbytenum_t chars);
        void appendLine();
//private:
        void recalcLines();
        sbytenum_t findLineStart(slinenum_t line);
};

class BufferMark;
class Buffer {
        Block *block;
        unsigned blocks;
        unsigned blocksAlloc;
        BufferMark *firstMark;
        
        bool changed;
        mutable llinenum_t cached_find_line;
        mutable llinenum_t cached_find_line_line;
        mutable unsigned cached_find_line_block;
        void invalidateCachedFindLine() const { cached_find_line=0; cached_find_line_line=0; cached_find_line_block=0; }
        
public:
        Buffer();
        virtual ~Buffer();
        
        enum { last_line=UINT_MAX, last_character=UINT_MAX };
        
        void splitLine(llinenum_t line, colnum_t col);
        void joinLine(llinenum_t line);
        void removeLine(llinenum_t line);
        void appendLine();
        
        void insert(llinenum_t line, colnum_t col, char c);
        void insert(llinenum_t line, colnum_t col, const char *s, int sl=-1);
        void remove(llinenum_t line, colnum_t col, unsigned chars);
        void clear();
        
        llinenum_t queryLineCount() const;
        char *queryLinePtr(llinenum_t line);
        const char *queryLinePtr(llinenum_t line) const;
        sbytenum_t queryLineLength(llinenum_t line) const;
        
        void addMark(BufferMark *bm);
        void removeMark(BufferMark *bm);
        
        void compact();
        void fastCompact();
        
        int queryChanged() const { return changed; }
        void setChanged(int newflag) { changed=newflag; }
private:
        unsigned findBlockWithLine(llinenum_t line, slinenum_t *rline) const;
        void splitBlock(unsigned &b, llinenum_t absoluteLine, slinenum_t &relativeLine);
        void insertBlock(unsigned before);
        void removeBlock(unsigned at);
        static void fixcolpo(const Block *bp, slinenum_t line, colnum_t &col);
public:
        void queryStatistics(lbytenum_t *used_bytes, lbytenum_t *allocated_bytes, llinenum_t *lines);
};


class BufferMark {
        llinenum_t line;
        colnum_t col;
        bool changed;
        BufferMark *next;
        friend class Buffer;
public:
        enum stickto { left,right};
private:
        stickto stick;
public:
        BufferMark()
          : line(0), col(0), changed(false), next(0), stick(right)
          {}
        BufferMark(llinenum_t l, colnum_t c, stickto s)
          : line(l), col(c), changed(false), next(0), stick(s)
          {}
        BufferMark(const BufferMark &bm)
          : line(bm.line), col(bm.col), changed(false), next(0), stick(bm.stick)
          {}
        virtual ~BufferMark() {};
        BufferMark &operator=(const BufferMark &bm) {
                line=bm.line; col=bm.col; changed=true;
                return *this;
        }
        
        llinenum_t queryLine() const { return line; }
        colnum_t queryCol() const { return col; }
        
        void setLine(llinenum_t nl) { line=nl; changed=true; }
        void setCol(colnum_t nc) { col=nc; changed=true; }
        void set(llinenum_t l, colnum_t c) { line=l; col=c; changed=1; }
        
        bool queryChanged() const { return changed; }
        void resetChanged() { changed=false; }
        
        int operator<(const BufferMark &bm) const 
          { return line<bm.line || (line==bm.line && col<bm.col); }
        int operator<=(const BufferMark &bm) const
          { return line<bm.line || (line==bm.line && col<=bm.col); }
        int operator==(const BufferMark &bm) const
          { return line==bm.line && col==bm.col; }
        int operator>=(const BufferMark &bm) const
          { return line>bm.line || (line==bm.line && col>=bm.col); }
        int operator>(const BufferMark &bm) const
          { return line>bm.line || (line==bm.line && col>bm.col); }
        int operator!=(const BufferMark &bm) const
          { return line!=bm.line || line!=bm.line; }
};

#endif /*_BUFFER_H_*/
